use std::rc::Rc;

use crate::{environment::Environment, node::function::FunctionType};

use super::{Node, NodeEnum, Precedence};

#[derive(Debug, Clone, PartialEq, PartialOrd)]
pub struct Call {
    function: Rc<NodeEnum>,
    arguments: Rc<NodeEnum>,
}

impl Node for Call {
    fn evaluate(&self, env: &mut Environment) -> Result<Rc<NodeEnum>, String> {
        // Evaluate callee and error if its not a function
        let evaluated = self.function.evaluate(env)?;
        let func = if let NodeEnum::Function(func) = evaluated.as_ref() {
            func
        } else {
            return Err(format!(
                "Cannot call {} as a function",
                evaluated.as_string(Some(env))
            ));
        };

        let arguments = if let NodeEnum::Set(set) = self.arguments.as_ref() {
            let mut arguments = vec![];
            for value in set.get_values().iter() {
                arguments.push(value.evaluate(env)?);
            }
            arguments
        } else {
            panic!(
                "This shoulnd not be possible, but call constructed with arguments of type that is not set"
            );
        };

        // Call function body with arguments
        match func.get_body() {
            // Pass arguments to native function
            FunctionType::Native(_name, native_function) => native_function(&arguments, env),
            FunctionType::UserFunction(body, fargs) => {
                if fargs.len() != arguments.len() {
                    return Err(format!(
                        "Error calling function. Expected {} arguments, but got {}",
                        fargs.len(),
                        arguments.len()
                    ));
                }
                env.push_stack()?;
                // Define variables
                fargs.iter().zip(arguments).for_each(|(symbol, value)| {
                    env.insert(symbol.get_value(), value.clone());
                });
                let ev = body.evaluate(env)?;
                env.pop_stack()?;
                // Return evaluated return value for function
                Ok(ev)
            }
        }
    }

    fn as_string(&self, env: Option<&Environment>) -> String {
        let arguments = self.arguments.as_string(env);
        format!("{}({})", self.function.as_string(env), arguments)
    }

    fn precedence(&self) -> Precedence {
        Precedence::Call
    }
}

impl Call {
    pub fn new(function: Rc<NodeEnum>, arguments: Rc<NodeEnum>) -> Rc<NodeEnum> {
        Rc::new(
            Self {
                function,
                arguments,
            }
            .into(),
        )
    }
}
